#  Copyright (c) 2013 Andrey B. Panfilov <andrew@panfilov.tel>
#
#  See main module for license.
#

from dctmpy import *
from dctmpy.obj.typedobject import TypedObject


class TypeObject(TypedObject):
    def __init__(self, **kwargs):
        self.__type_count = None
        super(TypeObject, self).__init__(**kwargs)

    def _read(self, buf=None):
        self._read_header()
        type_info = None
        if self.__type_count is not None:
            for i in xrange(0, self.__type_count):
                type_info = self._deserialize_child_type()
        else:
            while not isempty(self.buffer()):
                type_info = self._deserialize_child_type()
        self.type = type_info

    def _deserialize_child_type(self):
        child_type = self._read_type()
        if child_type is not None:
            add_type_to_cache(child_type)
        return child_type

    def _read_type(self):
        return super(TypeObject, self)._read_type()

    def _read_header(self):
        self.__type_count = self._read_int()
        if self.serversion > 0:
            self._read_int()

    def _need_read_type(self):
        return False

    def _need_read_object(self):
        return False



#  Copyright (c) 2013 Andrey B. Panfilov <andrew@panfilov.tel>
#
#  See main module for license.
#

from dctmpy.obj.typedobject import TypedObject


class EntryPoints(TypedObject):
    def __init__(self, **kwargs):
        self.__methods = None
        super(EntryPoints, self).__init__(**dict(
            kwargs,
            **{'serversion': 0}
        ))

    def _read(self, buf=None):
        super(EntryPoints, self)._read(buf)
        if len(self) > 0:
            self.__methods = dict(zip(self['name'], self['pos']))

    def methods(self):
        return self.__methods

    def __getattr__(self, name):
        if self.__methods:
            return self.__methods[name]
        else:
            return super(EntryPoints, self).__getattr__(name)

    def __setattr__(self, name, value):
        super(EntryPoints, self).__setattr__(name, value)





#  Copyright (c) 2013 Andrey B. Panfilov <andrew@panfilov.tel>
#
#  See main module for license.
#

from dctmpy import *


class TypedObject(object):
    attributes = ['session', 'type', 'buffer', 'serversion', 'iso8601time']

    def __init__(self, **kwargs):
        for attribute in TypedObject.attributes:
            self.__setattr__(ATTRIBUTE_PREFIX + attribute, kwargs.pop(attribute, None))
        self.__attrs = {}

        if self.serversion is None:
            self.serversion = self.session.serversion

        if self.iso8601time is None:
            if self.serversion == 2:
                self.iso8601time = self.session.iso8601time
            else:
                self.iso8601time = False

        if not isempty(self.buffer):
            self._read()

    def _read(self, buf=None):
        if isempty(buf) and isempty(self.buffer):
            raise ParserException("Empty data")
        elif not isempty(buf):
            self.buffer = buf

        self._read_header()

        if self.type is None and self._need_read_type():
            self.type = self._read_type()

        if self._need_read_object():
            self._read_object()

    def _read_header(self):
        if self.serversion > 0:
            serversion = self._read_int()
            if serversion != self.serversion:
                raise RuntimeError(
                    "Invalid serialization version %d, expected %d" % (serversion, self.serversion))

    def _read_type(self):
        header = self._next_token()
        if header != "TYPE":
            raise ParserException("Invalid type header: %s" % header)

        type_info = self._read_type_info()
        for i in xrange(0, self._read_int()):
            type_info.append(self._read_attr_info())

        return type_info

    def _read_object(self):
        header = self._next_token()
        if "OBJ" != header:
            raise ParserException("Invalid header, expected OBJ, got: %s" % header)

        type_name = self._next_token()

        if not type_name:
            raise ParserException("Wrong type name")

        if self.serversion > 0:
            self._read_int()
            self._read_int()
            self._read_int()

        if self.type is None or type_name != self.type.name:
            raise ParserException("No type info for %s" % type_name)

        for i in xrange(0, self._read_int()):
            self._read_attr(i)

        self._read_extended_attr()

    def _read_attr(self, index):
        position = self._if_d6(self._read_base64_int)
        if position is None:
            position = index

        repeating = self.type.get(position).repeating
        attr_type = self.type.get(position).type

        if self.serversion == 2:
            repeating = self._next_string(REPEATING_PATTERN) == REPEATING
            attr_type = TYPES[self._read_int()]

        attr_name = self.type.get(position).name
        attr_length = self.type.get(position).length

        if not attr_type:
            raise ParserException("Unknown type")

        result = []

        if not repeating:
            result.append(self._read_attr_value(attr_type))
        else:
            for i in xrange(0, self._read_int()):
                result.append(self._read_attr_value(attr_type))

        self.add(AttrValue(**{
            'name': attr_name,
            'position': position,
            'type': attr_type,
            'length': attr_length,
            'values': result,
            'repeating': repeating,
        }))

    def add(self, value):
        self.__attrs[value.name] = value

    def _read_extended_attr(self):
        attr_count = self._read_int()
        for i in xrange(0, attr_count):
            attr_name = self._next_string(ATTRIBUTE_PATTERN)
            attr_type = self._next_string(ATTRIBUTE_PATTERN)
            repeating = REPEATING == self._next_string()
            length = self._read_int()

            if isempty(attr_type):
                raise ParserException("Unknown typedef: %s" % attr_type)

            result = []

            if not repeating:
                result.append(self._read_attr_value(attr_type))
            else:
                for i in xrange(1, self._read_int()):
                    result.append(self._read_attr_value(attr_type))

            self.__attrs[attr_name] = AttrValue(**{
                'name': attr_name,
                'type': attr_type,
                'length': length,
                'values': result,
                'repeating': repeating,
            })

    def _read_attr_value(self, attr_type):
        return {
            INT: lambda: self._read_int(),
            STRING: lambda: self._read_string(),
            TIME: lambda: self._read_time(),
            BOOL: lambda: self._read_boolean(),
            ID: lambda: self._next_string(),
            DOUBLE: lambda: self._next_string(),
            UNDEFINED: lambda: self._next_string()
        }[attr_type]()

    def _read_type_info(self):
        return TypeInfo(**{
            'name': self._next_string(ATTRIBUTE_PATTERN),
            'id': self._next_string(ATTRIBUTE_PATTERN),
            'vstamp': self._if_d6(self._read_int),
            'version': self._if_d6(self._read_int),
            'cache': self._if_d6(self._read_int),
            'super': self._next_string(ATTRIBUTE_PATTERN),
            'sharedparent': self._if_d6(self._next_string, None, ATTRIBUTE_PATTERN),
            'aspectname': self._if_d6(self._next_string, None, ATTRIBUTE_PATTERN),
            'aspectshareflag': self._if_d6(self._read_boolean),
            'serversion': self.serversion,
        })

    def _read_attr_info(self):
        return AttrInfo(**{
            'position': self._if_d6(self._read_base64_int),
            'name': self._next_string(ATTRIBUTE_PATTERN),
            'type': self._next_string(TYPE_PATTERN),
            'repeating': REPEATING == self._next_string(),
            'length': self._read_int(),
            'restriction': self._if_d6(self._read_int),
        })

    def _if_d6(self, method, default=None, *args, **kwargs):
        if self.serversion > 0:
            return method(*args, **kwargs)
        return default

    def serialize(self):
        result = ""
        if self.serversion > 0:
            result += "%d\n" % self.serversion
        result += "OBJ NULL 0 "
        if self.serversion > 0:
            result += "0 0\n0\n"
        result += "%d\n" % len(self.__attrs)
        for attr_value in self.__attrs.values():
            result += "%s %s %s %d\n" % (
                attr_value.name, attr_value.type, [SINGLE, REPEATING][attr_value.repeating],
                attr_value.length)
            if attr_value.repeating:
                result += "%d\n" % len(attr_value.values)
            for value in attr_value.values:
                if STRING == attr_value.type:
                    result += "A %d %s\n" % (len(value), value)
                elif BOOL == attr_value.type:
                    result += "%s\n" % ["F", "T"][value]
                else:
                    result += "%s\n" % value
        return result

    def _need_read_type(self):
        return True

    def _need_read_object(self):
        return True

    def _substr(self, length):
        data = self.buffer
        self.buffer = data[length:]
        return data[:length]

    def _next_token(self, separator=DEFAULT_SEPARATOR):
        m = re.search(r'^%s' % separator, self.buffer)
        if m:
            self._substr(m.end(0))
        m = re.search(separator, self.buffer)
        if m:
            return self._substr(m.start(0))
        else:
            return self._substr(len(self.buffer))

    def _next_string(self, pattern=None, separator=DEFAULT_SEPARATOR):
        value = self._next_token(separator)
        if pattern:
            if not re.match(pattern, value):
                raise ParserException("Invalid string: %s for regexp %s" % (value, pattern))
        return value

    def _read_int(self):
        return int(self._next_string(INTEGER_PATTERN))

    def _read_base64_int(self):
        return pseudo_base64_to_int(self._next_string(BASE64_PATTERN))

    def _read_string(self):
        self._next_string(ENCODING_PATTERN)
        return self._substr(self._read_int() + 1)[1:]

    def _read_time(self):
        value = self._next_token(CRLF_PATTERN)
        if value.startswith(" "):
            value = value[1:]
        if value.startswith("xxx "):
            value = value[4:]
        return parse_time(value)

    def _read_boolean(self):
        value = self._next_string(BOOLEAN_PATTERN)
        return value == 'T' or value == '1'

    def __getattr__(self, name):
        if name in self.__attrs:
            return self.__attrs[name]
        elif name in TypedObject.attributes:
            return self.__getattribute__(ATTRIBUTE_PREFIX + name)
        else:
            raise AttributeError("Unknown attribute %s in %s" % (name, str(self.__class__)))

    def __setattr__(self, name, value):
        if name in TypedObject.attributes:
            TypedObject.__setattr__(self, ATTRIBUTE_PREFIX + name, value)
        else:
            super(TypedObject, self).__setattr__(name, value)

    def __len__(self):
        return len(self.__attrs)

    def __contains__(self, key):
        return key in self.__attrs

    def __getitem__(self, key):
        if key in self.__attrs:
            attr_value = self.__attrs[key]
            if attr_value.repeating:
                return attr_value.values
            else:
                return attr_value[0]
        else:
            raise KeyError

    def __setitem__(self, key, value):
        if key in self.__attrs:
            attr_value = self.__attrs[key]
            if attr_value.repeating:
                if value is None:
                    attr_value.values = []
                elif isinstance(value, list):
                    attr_value.values = value
                else:
                    attr_value.values = [value]
            else:
                if value is None:
                    attr_value.values = []
                elif isinstance(value, list):
                    if len(value) > 1:
                        raise RuntimeError("Single attribute %s does not accept arrays" % key)
                    elif len(value) == 0:
                        attr_value.values = []
                    else:
                        val = value[0]
                        if val is None:
                            attr_value.values = []
                        else:
                            attr_value.values = [val]
        else:
            raise KeyError

    def __iter__(self):
        return iter(self.__attrs.keys())



#  Copyright (c) 2013 Andrey B. Panfilov <andrew@panfilov.tel>
#
#  See main module for license.
#

import array
from dctmpy.net import *


class Response(object):
    def __init__(self, **kwargs):
        message = kwargs.pop('message', None)

        if message is None:
            raise ProtocolException("Response undefined")

        self.__message = array.array('B')
        self.__message.extend(message)
        self.__data = []

        self._deserialize()

    def _deserialize(self):
        while len(self.__message) > 0:
            if self.__message[0] == INTEGER_START:
                self.__data.append(read_integer(self.__message))
            elif self.__message[0] == EMPTY_STRING_START and self.__message[0] == NULL_BYTE:
                self.__data.append(read_string(self.__message))
            elif self.__message[0] == STRING_START:
                self.__data.append(read_string(self.__message))
            elif self.__message[0] == STRING_ARRAY_START and self.__message[1] == 0x80:
                self.__data.append(read_string(self.__message))
            elif self.__message[0] == INT_ARRAY_START:
                self.__data.append(read_integer_array(self.__message))

    def next(self):
        if len(self.__data) > 0:
            return self.__data.pop(0)
        return None

    def last(self):
        if len(self.__data) > 0:
            return self.__data.pop()
        return None


#  Copyright (c) 2013 Andrey B. Panfilov <andrew@panfilov.tel>
#
#  See main module for license.
#
import logging

from dctmpy import *
from dctmpy.netwise import Netwise
from dctmpy.obj.collection import Collection, PersistentCollection
from dctmpy.obj.entrypoints import EntryPoints
from dctmpy.obj.persistent import Persistent
from dctmpy.obj.type import TypeObject
from dctmpy.obj.typedobject import TypedObject

NETWISE_VERSION = 3
NETWISE_RELEASE = 5
NETWISE_INUMBER = 769


class DocbaseClient(Netwise):
    attributes = ['docbaseid', 'username', 'password', 'messages', 'entrypoints', 'serversion', 'iso8601time',
                  'session', 'serversionhint', 'docbaseconfig', 'serverconfg', 'readingmessages', 'knowncommands']

    def __init__(self, **kwargs):
        for attribute in DocbaseClient.attributes:
            self.__setattr__(ATTRIBUTE_PREFIX + attribute, kwargs.pop(attribute, None))
        super(DocbaseClient, self).__init__(**dict(kwargs, **{
            'version': NETWISE_VERSION,
            'release': NETWISE_RELEASE,
            'inumber': NETWISE_INUMBER,
        }))

        if self.serversion is None:
            self.serversion = 0
        if self.iso8601time is None:
            self.iso8601time = False
        if not self.session:
            self.session = NULL_ID
        if not self.docbaseid >= 0:
            self._resolve_docbase_id()
        if self.messages is None:
            self.messages = []
        if self.entrypoints is None:
            self.entrypoints = {
                'ENTRY_POINTS': 0,
                'GET_ERRORS': 558,
            }

        self._register_known_commands()

        for name in self.entrypoints.keys():
            self._add_entry_point(name)
        if self.serversionhint is None:
            self.serversionhint = CLIENT_VERSION_ARRAY[3]

        self.readingmessages = False

        self._connect()
        self._fetch_entry_points()

        if self.password and self.username:
            self.authenticate()

    def _resolve_docbase_id(self):
        response = self.request(
            type=RPC_NEW_SESSION_BY_ADDR,
            data=[
                -1,
                EMPTY_STRING,
                CLIENT_VERSION_STRING,
                EMPTY_STRING,
                CLIENT_VERSION_ARRAY,
                NULL_ID,
            ],
            immediate=True,
        ).receive()

        reason = response.next()
        m = re.search('Wrong docbase id: \(-1\) expecting: \((\d+)\)', reason)
        if m:
            self.docbaseid = int(m.group(1))
        self.disconnect()

    def disconnect(self):
        try:
            if self.session and self.session != NULL_ID:
                self.request(
                    type=RPC_CLOSE_SESSION,
                    data=[
                        self.session,
                    ],
                    immediate=True,
                ).receive()
            super(DocbaseClient, self).disconnect()
        finally:
            self.session = None

    def _reconnect(self):
        ''

    def _connect(self):
        response = self.request(
            type=RPC_NEW_SESSION_BY_ADDR,
            data=[
                self.docbaseid,
                EMPTY_STRING,
                CLIENT_VERSION_STRING,
                EMPTY_STRING,
                CLIENT_VERSION_ARRAY,
                NULL_ID,
            ],
            immediate=True,
        ).receive()

        reason = response.next()
        server_version = response.next()
        if server_version[7] == DM_CLIENT_SERIALIZATION_VERSION_HINT:
            self.serversion = DM_CLIENT_SERIALIZATION_VERSION_HINT
        else:
            self.serversion = 0

        if self.serversion == 0 or self.serversion == 1:
            self.iso8601time = False
        else:
            if server_version[9] & 0x01 != 0:
                self.iso8601time = False
            else:
                self.iso8601time = True

        session = response.next()

        if session == NULL_ID:
            raise RuntimeError(reason)

        self.session = session

    def rpc(self, rpc_id, data=None):
        if not data:
            data = []
        if self.session:
            if len(data) == 0 or data[0] != self.session:
                data.insert(0, self.session)

        (valid, o_data, collection, persistent, maybemore, records) = (None, None, None, None, None, None)

        response = self.request(type=rpc_id, data=data, immediate=True).receive()
        message = response.next()
        o_data = response.last()
        if rpc_id == RPC_APPLY_FOR_OBJECT:
            valid = int(response.next()) > 0
            persistent = int(response.next()) > 0
        elif rpc_id == RPC_APPLY:
            collection = int(response.next())
            persistent = int(response.next()) > 0
            maybemore = int(response.next()) > 0
            valid = collection >= 0
        elif rpc_id == RPC_CLOSE_COLLECTION:
            pass
        elif rpc_id == RPC_GET_NEXT_PIECE:
            pass
        elif rpc_id == RPC_MULTI_NEXT:
            records = int(response.next())
            maybemore = int(response.next()) > 0
            valid = int(response.next()) > 0
        else:
            valid = int(response.next()) > 0

        if (o_data & 0x02 != 0) and not self.readingmessages:
            try:
                self.readingmessages = True
                self._get_messages()
            finally:
                self.readingmessages = False

        #TODO in some cases (e.g. AUTHENTICATE_USER) CS returns both OBDATA and RESULT
        if o_data & 0x02 != 0 and len(self.messages) > 0:
            reason = self._get_message(3)
            if len(reason) > 0:
                raise RuntimeError(reason)
        elif valid is not None and not valid:
            raise RuntimeError("Unknown error")
        elif len(self.messages) > 0:
            logging.debug(self._get_message(0))

        if o_data == 0x10 or (o_data == 0x01 and rpc_id == RPC_GET_NEXT_PIECE):
            message += self.rpc(RPC_GET_NEXT_PIECE).data

        return Response(data=message, odata=o_data, persistent=persistent, collection=collection, maybemore=maybemore,
                        records=records)

    def apply(self, rpc_id, object_id, method, request=None, cls=Collection):
        if rpc_id is None:
            rpc_id = RPC_APPLY

        if not object_id:
            object_id = NULL_ID

        response = self.rpc(rpc_id, [self._get_method(method), object_id, request])
        data = response.data

        if rpc_id == RPC_APPLY_FOR_STRING:
            return data
        elif rpc_id == RPC_APPLY_FOR_ID:
            return data
        elif rpc_id == RPC_APPLY_FOR_DOUBLE:
            return data
        elif rpc_id == RPC_APPLY_FOR_BOOL:
            return data
        elif rpc_id == RPC_APPLY_FOR_LONG:
            return data
        elif rpc_id == RPC_APPLY_FOR_TIME:
            return data

        if not cls:
            if rpc_id == RPC_APPLY:
                cls = Collection
            elif rpc_id == RPC_APPLY_FOR_OBJECT:
                cls = TypedObject

        if response.persistent:
            if cls == Collection:
                cls = PersistentCollection
            elif cls == TypedObject:
                cls = Persistent

        result = cls(session=self, buffer=data)
        if response.collecton is not None and isinstance(result, Collection):
            result.collection = response.collection
            result.persistent = response.persistent
            result.records = response.records
            result.maybemore = response.maybemore
            if isinstance(request, TypedObject) and 'BATCH_HINT' in request:
                result.batchsize = request['BATCH_HINT']
            else:
                result.batchsize = DEFAULT_BATCH_SIZE

        return result

    def _get_messages(self):
        self.messages = [x for x in self.get_errors()]

    def _get_message(self, severity=0):
        if not self.messages:
            return ""
        message = ""
        for i in xrange(0, len(self.messages)):
            if self.messages[i]['SEVERITY'] < severity:
                continue
            if len(message) > 0:
                message += ", "
            message += "%s: %s" % (self.messages[i]['NAME'], self.messages[i]['1'])

        for i in xrange(len(self.messages) - 1, 1):
            if self.messages[i]['SEVERITY'] >= severity:
                self.messages.pop(i)

        return message

    def authenticate(self, username=None, password=None):
        if username and password:
            self.username = username
            self.password = password
        if not self.username:
            raise RuntimeError("Empty username")
        if not self.password:
            raise RuntimeError("Empty password")

        result = self.authenticate_user(self.username, self._obfuscate(self.password))
        if result['RETURN_VALUE'] != 1:
            raise RuntimeError("Unable to authenticate")

        self.docbaseconfig = self.get_docbase_config()
        self.serverconfig = self.get_server_config()

    def next_batch(self, collection, batch_hint=DEFAULT_BATCH_SIZE):
        return self.rpc(RPC_MULTI_NEXT, [collection, batch_hint])

    def close_collection(self, collection):
        self.rpc(RPC_CLOSE_COLLECTION, [collection])

    def _fetch_entry_points(self):
        self.entrypoints = self.entry_points().methods()
        for name in self.entrypoints:
            self._add_entry_point(name)

    def get_by_qualification(self, qualification):
        collection = self.query("select r_object_id from %s" % qualification)
        record = collection.next_record()
        if record:
            return self.fetch(record['r_object_id'])
        return None

    def get_type(self, name, vstamp=0):
        type_obj = get_type_from_cache(name)
        if type_obj:
            return type_obj
        data = None
        if "FETCH_TYPE" in self.entrypoints:
            data = self.fetch_type(name, vstamp)['result']
        else:
            data = self.rpc(RPC_FETCH_TYPE, [name]).data
        return TypeObject(session=self, buffer=data).type

    def query(self, query, for_update=False, batch_hint=DEFAULT_BATCH_SIZE, bof_dql=False):
        try:
            collection = self.execute(query, for_update, batch_hint, bof_dql)
        except Exception, e:
            raise RuntimeError("Error occurred while executing query: %s" % query, e)
        return collection

    def _obfuscate(self, password):
        if self._isobfuscated(password):
            return password
        return "".join(
            "%02x" % [x ^ 0xB6, 0xB6][x == 0xB6] for x in (ord(x) for x in password[::-1])
        )

    def _isobfuscated(self, password):
        if not re.match("^([0-9a-f]{2})+$", password):
            return False
        for x in re.findall("[0-9a-f]{2}", password):
            if int(x, 16) != 0xB6 and (int(x, 16) ^ 0xB6) > 127:
                return False
        return True

    def _as_object(self, object_id, method, request=None, cls=TypedObject):
        if not object_id:
            object_id = NULL_ID
        return self.apply(RPC_APPLY_FOR_OBJECT, object_id, method, request, cls)

    def _as_collection(self, object_id, method, request=None, cls=Collection):
        if not object_id:
            object_id = NULL_ID
        return self.apply(RPC_APPLY, object_id, method, request, cls)

    def _as_string(self, object_id, method, request=None, cls=None):
        if not object_id:
            object_id = NULL_ID
        return self.apply(RPC_APPLY_FOR_STRING, object_id, method, request, cls)

    def _as_id(self, object_id, method, request=None, cls=None):
        if not object_id:
            object_id = NULL_ID
        return self.apply(RPC_APPLY_FOR_ID, object_id, method, request, cls)

    def _as_time(self, object_id, method, request=None, cls=None):
        if not object_id:
            object_id = NULL_ID
        return self.apply(RPC_APPLY_FOR_TIME, object_id, method, request, cls)

    def _get_method(self, name):
        if name not in self.entrypoints:
            raise RuntimeError("Unknown method: %s" % name)
        return self.entrypoints[name]

    def __getattr__(self, name):
        if name in DocbaseClient.attributes:
            return self.__getattribute__(ATTRIBUTE_PREFIX + name)
        else:
            return super(DocbaseClient, self).__getattr__(name)

    def __setattr__(self, name, value):
        if name in DocbaseClient.attributes:
            DocbaseClient.__setattr__(self, ATTRIBUTE_PREFIX + name, value)
        else:
            super(DocbaseClient, self).__setattr__(name, value)

    def _add_entry_point(self, name):
        pep_name = DocbaseClient._pep_name(name)
        if getattr(DocbaseClient, pep_name, None):
            return
        elif name in self.knowncommands:
            command = self.knowncommands[name]
            method = command.method
            cls = command.returntype
            request = command.request
            needid = command.needid
            argc = command.argcnt
            if needid:
                def inner(self, object_id=NULL_ID, *args):
                    if not request:
                        return method(self, object_id, name, None, cls)
                    if argc == 1:
                        return method(self, object_id, name, request(self), cls)
                    else:
                        return method(self, object_id, name, request(self, *args), cls)
            else:
                def inner(self, *args):
                    if not request:
                        return method(self, NULL_ID, name, None, cls)
                    if argc == 1:
                        return method(self, NULL_ID, name, request(self), cls)
                    else:
                        return method(self, NULL_ID, name, request(self, *args), cls)
        else:
            def inner(self, object_id=NULL_ID, request=None, cls=Collection):
                return self._as_collection(object_id, name, request, cls)

        inner.__name__ = pep_name
        setattr(self.__class__, inner.__name__, inner)

    def _register_known_commands(self):
        cls = self.__class__
        self._register(Rpc('GET_SERVER_CONFIG', cls._as_object, TypedObject, cls._server_config_request, False))
        self._register(Rpc('GET_DOCBASE_CONFIG', cls._as_object, TypedObject, cls._docbase_config_request, False))
        self._register(Rpc('ENTRY_POINTS', cls._as_object, EntryPoints, cls._entry_points_request, False))
        self._register(Rpc('FETCH', cls._as_object, Persistent, None, True))
        self._register(Rpc('AUTHENTICATE_USER', cls._as_object, TypedObject, cls._auth_request, False))
        self._register(Rpc('GET_ERRORS', cls._as_collection, Collection, cls._get_errors_request, False))
        self._register(Rpc('FETCH_TYPE', cls._as_object, TypedObject, cls._fetch_type_request, False))
        self._register(Rpc('EXEC', cls._as_collection, Collection, cls._query_request, False))
        self._register(Rpc('TIME', cls._as_time, TypedObject, None, False))
        self._register(Rpc('COUNT_SESSIONS', cls._as_object, TypedObject, None, False))
        self._register(Rpc('EXEC_SELECT_SQL', cls._as_collection, Collection, cls._sql_query_request, False))
        self._register(Rpc('FTINDEX_AGENT_ADMIN', cls._as_object, TypedObject, cls._index_agent_status_request, False))
        self._register(Rpc('DUMP_JMS_CONFIG_LIST', cls._as_object, TypedObject, None, False))

    def _register(self, command):
        if not self.knowncommands:
            self.knowncommands = {}
        self.knowncommands[command.command] = command

    @staticmethod
    def _entry_points_request(session):
        obj = TypedObject(session=session)
        obj.add(AttrValue(name="LANGUAGE", type=INT, values=[get_locale_id()]))
        obj.add(AttrValue(name="CHARACTER_SET", type=INT, values=[get_charset_id()]))
        obj.add(AttrValue(name="PLATFORM_ENUM", type=INT, values=[get_platform_id()]))
        obj.add(AttrValue(name="PLATFORM_VERSION_IMAGE", type=STRING, values=["python"]))
        obj.add(AttrValue(name="UTC_OFFSET", type=INT, values=[get_offset_in_seconds()]))
        obj.add(AttrValue(name="SDF_AN_custom_date_order", type=INT, values=[0]))
        obj.add(AttrValue(name="SDF_AN_custom_scan_fields", type=INT, values=[0]))
        obj.add(AttrValue(name="SDF_AN_date_separator", type=STRING, values=["/"]))
        obj.add(AttrValue(name="SDF_AN_date_order", type=INT, values=[2]))
        obj.add(AttrValue(name="SDF_AN_day_leading_zero", type=BOOL, values=[True]))
        obj.add(AttrValue(name="SDF_AN_month_leading_zero", type=BOOL, values=[True]))
        obj.add(AttrValue(name="SDF_AN_century", type=BOOL, values=[True]))
        obj.add(AttrValue(name="SDF_AN_time_separator", type=STRING, values=[":"]))
        obj.add(AttrValue(name="SDF_AN_hours_24", type=BOOL, values=[True]))
        obj.add(AttrValue(name="SDF_AN_hour_leading_zero", type=BOOL, values=[True]))
        obj.add(AttrValue(name="SDF_AN_noon_is_zero", type=BOOL, values=[False]))
        obj.add(AttrValue(name="SDF_AN_am", type=STRING, values=["AM"]))
        obj.add(AttrValue(name="SDF_AN_pm", type=STRING, values=["PM"]))
        obj.add(AttrValue(name="PLATFORM_EXTRA", type=INT, repeating=True, values=[0, 0, 0, 0]))
        obj.add(AttrValue(name="APPLICATION_CODE", type=STRING, values=[""]))
        return obj

    @staticmethod
    def _auth_request(session, username, password):
        obj = TypedObject(session=session)
        obj.add(AttrValue(name="CONNECT_POOLING", type=BOOL, values=[False]))
        obj.add(AttrValue(name="USER_PASSWORD", type=STRING, values=[password]))
        obj.add(AttrValue(name="AUTHENTICATION_ONLY", type=BOOL, values=[False]))
        obj.add(AttrValue(name="CHECK_ONLY", type=BOOL, values=[False]))
        obj.add(AttrValue(name="LOGON_NAME", type=STRING, values=[username]))
        return obj

    @staticmethod
    def _server_config_request(session):
        obj = TypedObject(session=session)
        obj.add(AttrValue(name="OBJECT_TYPE", type=STRING, values=["dm_server_config"]))
        obj.add(AttrValue(name="FOR_REVERT", type=BOOL, values=[False]))
        obj.add(AttrValue(name="CACHE_VSTAMP", type=INT, values=[0]))
        return obj

    @staticmethod
    def _docbase_config_request(session):
        obj = TypedObject(session=session)
        obj.add(AttrValue(name="OBJECT_TYPE", type=STRING, values=["dm_docbase_config"]))
        obj.add(AttrValue(name="FOR_REVERT", type=BOOL, values=[False]))
        obj.add(AttrValue(name="CACHE_VSTAMP", type=INT, values=[0]))
        return obj

    @staticmethod
    def _fetch_type_request(session, typename, vstamp):
        obj = TypedObject(session=session)
        obj.add(AttrValue(name="TYPE_NAME", type=STRING, values=[typename]))
        obj.add(AttrValue(name="CACHE_VSTAMP", type=INT, values=[vstamp]))
        return obj

    @staticmethod
    def _get_errors_request(session):
        obj = TypedObject(session=session)
        obj.add(AttrValue(name="OBJECT_TYPE", type=STRING, values=["dmError"]))
        return obj

    @staticmethod
    def _query_request(session, query, for_update, batch_hint, bof_dql=False):
        obj = TypedObject(session=session)
        obj.add(AttrValue(name="QUERY", type=STRING, values=[query]))
        obj.add(AttrValue(name="FOR_UPDATE", type=BOOL, values=[for_update]))
        obj.add(AttrValue(name="BATCH_HINT", type=INT, values=[batch_hint]))
        obj.add(AttrValue(name="BOF_DQL", type=BOOL, values=[bof_dql]))
        return obj

    @staticmethod
    def _sql_query_request(session, query, batch_hint):
        obj = TypedObject(session=session)
        obj.add(AttrValue(name="QUERY", type=STRING, values=[query]))
        obj.add(AttrValue(name="BATCH_HINT", type=INT, values=[batch_hint]))
        return obj

    @staticmethod
    def _folder_by_path_request(session, path):
        obj = TypedObject(session=session)
        obj.add(AttrValue(name="_FOLDER_PATH_", type=STRING, values=[path]))
        return obj

    @staticmethod
    def _index_agent_status_request(session, indexname, agentname):
        obj = TypedObject(session=session)
        obj.add(AttrValue(name="NAME", type=STRING, values=[indexname]))
        obj.add(AttrValue(name="AGENT_INSTANCE_NAME", type=STRING, values=[agentname]))
        obj.add(AttrValue(name="ACTION", type=STRING, values=["status"]))
        return obj

    @staticmethod
    def _pep_name(rpc_name):
        if 'EXEC' == rpc_name:
            return "execute"
        s = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', rpc_name)
        return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s).lower()


class Response(object):
    attributes = ['data', 'odata', 'persistent', 'collection', 'records', 'maybemore']

    def __init__(self, **kwargs):
        for attribute in Response.attributes:
            self.__setattr__(ATTRIBUTE_PREFIX + attribute, kwargs.pop(attribute, None))

    def __getattr__(self, name):
        if name in Response.attributes:
            return self.__getattribute__(ATTRIBUTE_PREFIX + name)
        else:
            return AttributeError("Unknown attribute %s in %s" % (name, str(self.__class__)))

    def __setattr__(self, name, value):
        if name in Response.attributes:
            Response.__setattr__(self, ATTRIBUTE_PREFIX + name, value)
        else:
            super(Response, self).__setattr__(name, value)


class Rpc(object):
    attributes = ['command', 'method', 'returntype', 'request', 'needid', 'argcnt']

    def __init__(self, command, method, returntype, request, needid):
        self.command = command
        self.method = method
        self.returntype = returntype
        self.request = request
        self.needid = needid
        if self.request:
            self.argcnt = self.request.func_code.co_argcount
        else:
            self.argcnt = 0

    def __getattr__(self, name):
        if name in Rpc.attributes:
            return self.__getattribute__(ATTRIBUTE_PREFIX + name)
        else:
            return AttributeError("Unknown attribute %s in %s" % (name, str(self.__class__)))

    def __setattr__(self, name, value):
        if name in Rpc.attributes:
            Rpc.__setattr__(self, ATTRIBUTE_PREFIX + name, value)
        else:
            super(Rpc, self).__setattr__(name, value)





#  Copyright (c) 2013 Andrey B. Panfilov <andrew@panfilov.tel>
#
#  See main module for license.
#

import socket

from dctmpy import *
from dctmpy.net.request import Request


class Netwise(object):
    attributes = ['version', 'release', 'inumber', 'sequence', 'sockopts']

    def __init__(self, **kwargs):
        for attribute in Netwise.attributes:
            self.__setattr__(ATTRIBUTE_PREFIX + attribute, kwargs.pop(attribute, None))
        if self.sockopts is None:
            self.sockopts = kwargs
        if self.sequence is None:
            self.sequence = 0
        self.__socket = None

    def _connected(self):
        if not self.__socket:
            return False
        return True

    def _socket(self):
        if not self._connected():
            try:
                host = self.sockopts.get('host', None)
                port = self.sockopts.get('port', None)
                if not host or not (port > -1):
                    raise RuntimeError("Invalid host or port")
                self.__socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                self.__socket.connect((host, port))
            except:
                self.__socket = None
                raise
        return self.__socket

    def disconnect(self):
        try:
            if self._connected():
                self.__socket.close()
        finally:
            self.__socket = None

    def __del__(self):
        self.disconnect()

    def request(self, **kwargs):
        return Request(**dict(kwargs, **{
            'socket': self._socket(),
            'sequence': ++self.sequence,
            'version': self.version,
            'release': self.release,
            'inumber': self.inumber,
        }))

    def __getattr__(self, name):
        if name in Netwise.attributes:
            return self.__getattribute__(ATTRIBUTE_PREFIX + name)
        else:
            raise AttributeError("Unknown attribute %s in %s" % (name, str(self.__class__)))

    def __setattr__(self, name, value):
        if name in Netwise.attributes:
            Netwise.__setattr__(self, ATTRIBUTE_PREFIX + name, value)
        else:
            super(Netwise, self).__setattr__(name, value)




#  Copyright (c) 2013 Andrey B. Panfilov <andrew@panfilov.tel>
#
#  See main module for license.
#

from dctmpy import *
from dctmpy.obj.typedobject import TypedObject

PROTOCOL_VERSION = 0x30
INTEGER_START = 0x02
EMPTY_STRING_START = 0x05
NULL_BYTE = 0x00
STRING_START = 0x16
INT_ARRAY_START = 0x30
LONG_LENGTH_START = 0x82
STRING_ARRAY_START = 0x36


def serialize_integer(value):
    if value is None:
        raise RuntimeError("Undefined integer value")
    result = array.array('B')
    while value > 127 or value < -128:
        result.insert(0, value & 0x000000ff)
        value >>= 8
    result.insert(0, value & 0xFF)
    result.insert(0, len(result))
    result.insert(0, INTEGER_START)
    return result


def serialize_length(value):
    if value is None:
        raise RuntimeError("Undefined integer value")
    result = array.array('B')
    while value > 127:
        result.insert(0, value & 0x000000ff)
        value >>= 8
    result.insert(0, value)
    if len(result) > 1:
        result.insert(0, len(result) | 0x80)
    return result


def serialize_string(value):
    result = array.array('B')
    if not value:
        result.append(EMPTY_STRING_START)
        result.append(NULL_BYTE)
        return result
    result.extend(string_to_integer_array(value))
    result.append(NULL_BYTE)
    for b in serialize_length(len(result))[::-1]:
        result.insert(0, b)
    result.insert(0, STRING_START)
    return result


def serialize_id(value):
    if not value:
        value = NULL_ID
    return serialize_string(value)


def serialize_integer_array(intarray):
    result = []
    for i in intarray:
        result.extend(serialize_integer(i))
    result.insert(0, len(result) & 0x000000ff)
    result.insert(0, (len(result) - 1) >> 8)
    result.insert(0, LONG_LENGTH_START)
    result.insert(0, INT_ARRAY_START)
    return result


def serialize_value(value):
    if value is None:
        return serialize_string("")
    elif isinstance(value, str):
        return serialize_string(value)
    elif isinstance(value, int):
        return serialize_integer(value)
    elif isinstance(value, list):
        return serialize_integer_array(value)
    elif isinstance(value, TypedObject):
        return serialize_string(value.serialize())
    elif hasattr(value.__class__, "serialize"):
        return serialize_string(value.serialize())
    else:
        raise TypeError("Invalid argument type")


def serialize_data(data=None):
    result = array.array('B')
    if data is None:
        return result
    for i in data:
        result.extend(serialize_value(i))
    return result


def read_integer(data):
    if len(data) < 3:
        raise RuntimeError("Wrong sequence, at least 3 bytes required, got: %d" % len(data))

    header = data.pop(0)
    if header != INTEGER_START:
        raise RuntimeError("Wrong sequence for integer: 0x%X" % header)

    length = data.pop(0)
    if len(data) < length:
        raise RuntimeError("Wrong sequence, at least %d bytes required, got: %d" % (length, len(data)))

    value = data.pop(0)
    value -= (value & 0x80) << 1
    for i in xrange(1, length):
        value = value << 8 | data.pop(0)
        if value > 0x7fffffff:
            value = value - 0xffffffff - 1
    return value


def read_length(data):
    if len(data) < 1:
        raise RuntimeError("Wrong sequence, at least 1 byte required, got: %d" % len(data))

    value = data.pop(0)

    if value <= 127:
        return value

    length = value & 0x7F

    if len(data) < length:
        raise RuntimeError("Wrong sequence, %d bytes required, got: %d" % (length, len(data)))

    value = data.pop(0)
    for i in xrange(1, length):
        value = value << 8 | data.pop(0)
        if value > 0x7fffffff:
            value = value - 0xffffffff - 1
    return value


def read_integer_array(data):
    header = data.pop(0)
    if header != INT_ARRAY_START:
        raise RuntimeError("Wrong sequence for integer array: 0x%X" % header)

    length = read_length(data)
    stop = len(data) - length
    result = []
    while len(data) > stop:
        result.append(read_integer(data))
    return result


def read_array(data, asstring=False):
    sequence = data.pop(0)
    if sequence == EMPTY_STRING_START and data.pop(0) == NULL_BYTE:
        return []
    elif sequence == STRING_START:
        length = read_length(data)
        result = []
        for i in xrange(0, length):
            result.append(data.pop(0))
        if asstring and result[len(result) - 1] == NULL_BYTE:
            result.pop()
        return result
    elif sequence == STRING_ARRAY_START and data.pop(0) == 0x80:
        result = []
        while data[0] != NULL_BYTE or data[1] != NULL_BYTE:
            result.append(read_array(data, asstring))
        data.pop(0)
        data.pop(0)
        return result
    raise RuntimeError("Unknown sequence: 0x%X" % sequence)


def read_string(data):
    return integer_array_to_string(read_array(data, True))


def read_strings(data):
    result = []
    for res in read_array(data):
        result.append(integer_array_to_string(res))
    return result




#  Copyright (c) 2013 Andrey B. Panfilov <andrew@panfilov.tel>
#
#  See main module for license.
#

from dctmpy import *
from dctmpy.obj.typedobject import TypedObject


class DocbrokerObject(TypedObject):
    def __init__(self, **kwargs):
        super(DocbrokerObject, self).__init__(**dict(
            kwargs,
            **{'serversion': 0}
        ))

    def _read_type(self):
        pass

    def _read_object(self):
        header = self._next_token()
        if "OBJ" != header:
            raise ParserException("Invalid header, expected OBJ, got: %s" % header)

        type_name = self._next_token()
        if isempty(type_name):
            raise ParserException("Wrong type name")

        self._read_int()

        for i in xrange(0, self._read_int()):
            self._read_attr(i)

    def _read_attr(self, index):
        attr_name = self._next_string(ATTRIBUTE_PATTERN)
        attr_type = self._next_string(ATTRIBUTE_PATTERN)
        repeating = self._next_string(REPEATING_PATTERN) == REPEATING
        attr_length = self._read_int()

        if attr_type is None:
            raise ParserException("Unknown type")

        result = []

        if not repeating:
            result.append(self._read_attr_value(attr_type))
        else:
            for i in xrange(0, self._read_int()):
                result.append(self._read_attr_value(attr_type))

        self.add(AttrValue(**{
            'name': attr_name,
            'type': attr_type,
            'length': attr_length,
            'values': result,
            'repeating': repeating,
        }))


class DocbaseMap(DocbrokerObject):
    def __init__(self, **kwargs):
        super(DocbaseMap, self).__init__(**kwargs)

    def add(self, value):
        if not value.name in ['i_host_addr']:
            value.repeating = True
        super(DocbrokerObject, self).add(value)



import sys

from setuptools import setup


if sys.version_info < (2, 7):
    extras_require = {'nagios': ['argparse', 'nagiosplugin']}
else:
    extras_require = {'nagios': ['nagiosplugin']}

setup(
    name='dctmpy',
    version='0.1.7',
    packages=['dctmpy', 'dctmpy.net', 'dctmpy.obj', 'dctmpy.nagios'],
    package_dir={'': 'src'},
    license='ZPL-2.1',
    url='https://github.com/andreybpanfilov/dctmpy',
    author='Andrey B. Panfilov',
    author_email='andrew@panfilov.tel',
    description='Python bindings for Documentum',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Plugins',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: Zope Public License',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: System :: Monitoring',
    ],
    zip_safe=False,
    extras_require=extras_require,
    entry_points={
        'console_scripts':
            ['nagios_check_docbase = dctmpy.nagios.check_docbase:main [nagios]',
             'nagios_check_docbroker = dctmpy.nagios.check_docbroker:main [nagios]']
    }
)



#  Copyright (c) 2013 Andrey B. Panfilov <andrew@panfilov.tel>
#
#  See main module for license.
#

from dctmpy import *
from dctmpy.obj.typedobject import TypedObject


class Collection(TypedObject):
    attributes = ['collection', 'batchsize', 'records', 'maybemore', 'persistent']

    def __init__(self, **kwargs):
        for attribute in Collection.attributes:
            self.__setattr__(ATTRIBUTE_PREFIX + attribute, kwargs.pop(attribute, None))
        super(Collection, self).__init__(**kwargs)

    def _need_read_type(self):
        return True

    def _need_read_object(self):
        return False

    def next_record(self):
        if self.collection is None:
            return None

        if isempty(self.buffer) and (self.maybemore is None or self.maybemore):
            response = self.session.next_batch(self.collection, self.batchsize)
            self.buffer = response.data
            self.records = response.records
            self.maybemore = response.maybemore
            if self.serversion > 0 and not isempty(self.buffer):
                self._read_int()

        if not isempty(self.buffer) and (self.records is None or self.records > 0):
            try:
                cls = [CollectionEntry, PersistentCollectionEntry][self.persistent]
                entry = cls(session=self.session, type=self.type, buffer=self.buffer)
                self.buffer = entry.buffer
                return entry
            finally:
                if self.records is not None:
                    self.records -= 1
        try:
            self.close()
        except Exception, e:
            pass
        return None

    def __iter__(self):
        class iterator(object):
            def __init__(self, obj):
                self.obj = obj
                self.index = -1

            def __iter__(self):
                return self

            def next(self):
                r = self.obj.next_record()
                if r is None:
                    raise StopIteration
                else:
                    return r

        return iterator(self)

    def __getattr__(self, name):
        if name in Collection.attributes:
            return self.__getattribute__(ATTRIBUTE_PREFIX + name)
        else:
            return super(Collection, self).__getattr__(name)

    def __setattr__(self, name, value):
        if name in Collection.attributes:
            Collection.__setattr__(self, ATTRIBUTE_PREFIX + name, value)
        else:
            super(Collection, self).__setattr__(name, value)

    def close(self):
        if self.collection >= 0:
            try:
                self.session.close_collection(self.collection)
            finally:
                self.collection = None

    def __del__(self):
        self.close()


class PersistentCollection(Collection):
    def __init__(self, **kwargs):
        super(PersistentCollection, self).__init__(**kwargs)

    def _read(self, buf=None):
        if isempty(buf) and isempty(self.buffer):
            raise ParserException("Empty data")
        if not isempty(buf):
            self.buffer = buf
        self.type = self.session.get_type(self._next_string(), 0)

    def _need_read_type(self):
        return False

    def next_record(self):
        return super(PersistentCollection, self).next_record()

    def __getattr__(self, name):
        return super(PersistentCollection, self).__getattr__(name)

    def __setattr__(self, name, value):
        super(PersistentCollection, self).__setattr__(name, value)


class CollectionEntry(TypedObject):
    def __init__(self, **kwargs):
        super(CollectionEntry, self).__init__(**kwargs)

    def _read_header(self):
        pass

    def _read(self, buf=None):
        super(CollectionEntry, self)._read(buf)
        if self.serversion > 0:
            self._read_int()

    def __getattr__(self, name):
        if name in CollectionEntry.attributes:
            return self.__getattribute__(ATTRIBUTE_PREFIX + name)
        else:
            return super(CollectionEntry, self).__getattr__(name)

    def __setattr__(self, name, value):
        if name in CollectionEntry.attributes:
            CollectionEntry.__setattr__(self, ATTRIBUTE_PREFIX + name, value)
        else:
            super(CollectionEntry, self).__setattr__(name, value)


class PersistentCollectionEntry(CollectionEntry):
    def __init__(self, **kwargs):
        super(PersistentCollectionEntry, self).__init__(**kwargs)

    def _read_header(self):
        if not self.serversion > 0:
            self._next_string()

    def _read(self, buf=None):
        super(PersistentCollectionEntry, self)._read(buf)
        if self.serversion > 0:
            self._read_int()

    def __getattr__(self, name):
        return super(PersistentCollectionEntry, self).__getattr__(name)

    def __setattr__(self, name, value):
        super(PersistentCollectionEntry, self).__setattr__(name, value)
